#include "postgres.h"

#include <ctype.h>
#include <limits.h>

#include "access/hash.h"
#include "access/xlog.h"
#include "catalog/pg_type.h"
#include "libpq/pqformat.h"
#include "miscadmin.h"
#include "utils/array.h"
#include "utils/builtins.h"

#include <lz4.h>
#include "pg_lz4.h"


PG_MODULE_MAGIC;

PG_FUNCTION_INFO_V1(lz4_compression_handler);

static void
lz4_configure(Form_pg_attribute attr, List *options)
{
	if (options != NIL)
		elog(ERROR, "the lz4 compression method doesn't take any options");
}

/*
 * Compress tsvector using LZ4 compression.
 */
static struct varlena *
lz4_compress(CompressionOptions * ac, const struct varlena *data)
{
	/* input */
	int32		raw_len = VARSIZE_ANY_EXHDR(data);
	char	   *raw_data = VARDATA_ANY(data);

	/* output */
	int32		out_len;
	char	   *out_data;

	int32		clen;

	Assert(!VARATT_IS_COMPRESSED(data));

	out_len = LZ4_compressBound(raw_len);
	out_data = palloc0(out_len + VARHDRSZ_CUSTOM_COMPRESSED);

	/* try the compression */
	clen = LZ4_compress_default(raw_data,
								out_data + VARHDRSZ_CUSTOM_COMPRESSED,
								raw_len,
								out_len);

	if (clen > 0)
	{
		SET_VARSIZE_COMPRESSED(out_data, clen + VARHDRSZ_CUSTOM_COMPRESSED);
		return (struct varlena *) out_data;
	}

	pfree(out_data);
	return NULL;
}

static struct varlena *
lz4_decompress(CompressionOptions * ac, const struct varlena *data)
{
	int			dlen;

	/* input */
	char	   *raw_data = (char *) data + VARHDRSZ_CUSTOM_COMPRESSED;
	int32		raw_len = VARSIZE(data) - VARHDRSZ_CUSTOM_COMPRESSED;

	/* output */
	char	   *out_data;
	int32		out_len = VARRAWSIZE_4B_C(data) + VARHDRSZ;

	Assert(VARATT_IS_CUSTOM_COMPRESSED(data));

	out_data = palloc0(out_len);
	SET_VARSIZE(out_data, out_len);

	dlen = LZ4_decompress_safe(raw_data, VARDATA(out_data), raw_len, out_len - VARHDRSZ);

	if (dlen + VARHDRSZ != out_len)
		elog(ERROR, "compressed value is corrupted (%d != %d)", dlen, out_len);

	return (struct varlena *) out_data;
}

Datum 
lz4_compression_handler(PG_FUNCTION_ARGS)
{
	CompressionMethodRoutine *cmr;

	cmr = makeNode(CompressionMethodRoutine);

	cmr->configure = lz4_configure;
	cmr->drop = NULL;
	cmr->compress = lz4_compress;
	cmr->decompress = lz4_decompress;

	PG_RETURN_POINTER(cmr);
}
